"""Authentication Provider Plugin
Allows simple authentication using Username-Password pair with password saved into
Flaschengeist database (as User attribute)
"""

import os
import hashlib
import binascii
from werkzeug.exceptions import BadRequest
from flaschengeist.plugins import AuthPlugin
from flaschengeist.models import User, Role, Permission
from flaschengeist.database import db
from flaschengeist import logger


class AuthPlain(AuthPlugin):
    def can_register(self):
        return True

    def login(self, login_name, password):
        users: list[User] = (
            User.query.filter((User.userid == login_name) | (User.mail == login_name))
            .filter(User._attributes.any(name="password"))
            .all()
        )
        for user in users:
            if AuthPlain._verify_password(user.get_attribute("password"), password):
                return user.userid
        return False

    def modify_user(self, user, password, new_password=None):
        if password is not None and not self.login(user, password):
            raise BadRequest
        if new_password:
            user.set_attribute("password", AuthPlain._hash_password(new_password))

    def user_exists(self, userid) -> bool:
        return (
            db.session.query(User.id_).filter(User.userid == userid, User._attributes.any(name="password")).first()
            is not None
        )

    def create_user(self, user, password):
        if not user.userid:
            raise BadRequest("userid is missing for new user")
        hashed = AuthPlain._hash_password(password)
        user.set_attribute("password", hashed)

    def delete_user(self, user):
        pass

    @staticmethod
    def _hash_password(password):
        salt = hashlib.sha256(os.urandom(60)).hexdigest().encode("ascii")
        pass_hash = hashlib.pbkdf2_hmac("sha3-512", password.encode("utf-8"), salt, 100000)
        pass_hash = binascii.hexlify(pass_hash)
        return (salt + pass_hash).decode("ascii")

    @staticmethod
    def _verify_password(stored_password: str, provided_password: str):
        salt = stored_password[:64]
        stored_password = stored_password[64:]
        pass_hash = hashlib.pbkdf2_hmac("sha3-512", provided_password.encode("utf-8"), salt.encode("ascii"), 100000)
        pass_hash = binascii.hexlify(pass_hash).decode("ascii")
        return pass_hash == stored_password
